/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: mathndsmain						*/
/*                                                         		*/
/* Module Description: implements the MathLink interface for NDS	*/
/*                                                         		*/
/*----------------------------------------------------------------------*/
#define _CONFIG_DYNAMIC

#include <math.h>
#include <vector>
#include <iostream>
extern "C" {
#include "mathlink.h"
}
#include "dtt/gdsutil.h"
#include "dtt/gdschannel.h"
#include "dtt/confinfo.h"
#include "DAQSocket.hh"
#include "preproc.h"


#define USE_N
#define _ONESEC			1000000000LL
#define __ONESEC		1E9
#define _KEEPALIVEINTERVAL	60
#define DAQD_PORT		8088

/* dataType Defintions */
#define DAQ_DATATYPE_16BIT_INT	1	/* Data type signed 16bit integer */
#define DAQ_DATATYPE_32BIT_INT	2	/* Data type signed 32bit integer */
#define DAQ_DATATYPE_64BIT_INT	3	/* Data type signed 64bit integer */
#define DAQ_DATATYPE_FLOAT	4	/* Data type 32bit floating point */
#define DAQ_DATATYPE_DOUBLE	5	/* Data type 64bit double float */
#define DAQ_DATATYPE_STRING	6	/* Data type 32 char string */

   using namespace diag;
   using namespace std;


/*----------------------------------------------------------------------*/
/*                                                         		*/
/* NDS functions							*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

   class ndschninfo {
   public:
      gdsChnInfo_t	fInfo;
      preproc_t*	fPre;
      int		ndata;
      int		complex;
      float*		x;
   
      ndschninfo (gdsChnInfo_t& info) {
         fInfo = info;
         fPre = 0;
      }
      ~ndschninfo() {
         if (fPre != 0) {
            freePreproc (fPre);
         }
      }
   };

   typedef vector<ndschninfo> ndschnlist;
   ndschnlist clist;
   DAQSocket nds;


   bool connectNDS () 
   {
      nds.close();
   
      char 	daqServer[256];
      int	daqPort = 0;
      int 	status;
   
      // get NDS parameters
      // dynamic configuration
   #if defined (_CONFIG_DYNAMIC)
      const char* const* cinfo;	// configuration info
      confinfo_t	crec;		// conf. info record
      for (cinfo = getConfInfo (0, 0); *cinfo != NULL; cinfo++) {
         if ((parseConfInfo (*cinfo, &crec) == 0) &&
            (gds_strcasecmp (crec.interface, 
                            CONFIG_SERVICE_NDS) == 0) &&
            (crec.ifo == -1) && (crec.progver == -1)) {
            strcpy (daqServer, crec.host);
            daqPort = crec.port_prognum;
         }
      }
   #else
      // from parameter file
      strcpy (daqServer, DAQD_SERVER);
      loadStringParam (PRM_FILE, PRM_SECTION, PRM_SERVERNAME, daqServer);
      daqPort = DAQD_PORT;
      loadIntParam (PRM_FILE, PRM_SECTION, PRM_SERVERPORT, &daqPort);
   #endif
      if (daqPort <= 0) {
         daqPort = DAQD_PORT;
      }
   
      // connect to NDS
      status = nds.open (daqServer, daqPort);
      return (status == 0);
   }


   bool addChannelsNDS (const char* channels, bool trend = false,
                        bool simpletrend = false) 
   {
      const char* 	p = channels;
      int		num = 0;
      char		name[256] = {0};
      gdsChnInfo_t	info;
   
      clist.clear();
      while ((p != 0) && (*p != 0)) {
         // get name
         const char* end = strstr (p, "\\011");
         if (end == 0) {
            strncpy (name, p, 255);
            p += strlen (p);
         }
         else {
            int len = min ((int)(end-p), 255);
            strncpy (name, p, len);
            name[len] = 0;
            p = end + 4;
         }
      
         // lookup channel
         if (gdsChannelInfo (name, &info) != 0) {
            return false;
         }
         if (!trend) {
            if (!nds.AddChannel (info.chName, DAQSocket::rate_bps_pair 
                                (info.dataRate, info.bps))) {
               return false;
            }
         }
         else {
            char	buf[256];
#ifdef USE_N
            strcpy (strecpy (buf, info.chName), ".n");
            if (!simpletrend && 
	        !nds.AddChannel (buf, DAQSocket::rate_bps_pair 
                                (info.dataRate, 4))) {
               return false;
            }
#endif
            strcpy (strecpy (buf, info.chName), ".mean");
            if (!nds.AddChannel (buf, DAQSocket::rate_bps_pair 
                                (info.dataRate, 8))) {
               return false;
            }
            strcpy (strecpy (buf, info.chName), ".min");
            if (!simpletrend && 
	        !nds.AddChannel (buf, DAQSocket::rate_bps_pair 
                                (info.dataRate, 4))) {
               return false;
            }
            strcpy (strecpy (buf, info.chName), ".max");
            if (!simpletrend && 
	        !nds.AddChannel (buf, DAQSocket::rate_bps_pair 
                                (info.dataRate, 4))) {
               return false;
            }
            strcpy (strecpy (buf, info.chName), ".rms");
            if (!simpletrend && 
	        !nds.AddChannel (buf, DAQSocket::rate_bps_pair 
                                (info.dataRate, 8))) {
               return false;
            }
         }
         clist.push_back (ndschninfo (info));
         num++;
      }
      return true;
   }


/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Interface: NDSGetChannel						*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

   static int maxChn;
   static int countChn;


extern "C" 
   int addChannelName (const gdsChnInfo_t* info)
   {
      if (countChn < maxChn) {
         /*printf ("%s\n", info->chName);*/
         MLPutString (stdlink, info->chName);
      }
      countChn++;
      return 1;
   }


extern "C" 
   void ndsgetchannels ()
   {
      int len = gdsChannelListLen (-1, 0);
      MLPutFunction (stdlink, "List", len);
      maxChn = len;
      countChn = 0;
      gdsChannelListLen (-1, addChannelName);
   }


/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Interface: NDSGetChannelInfo						*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

extern "C" 
   int addChannelInfo (const gdsChnInfo_t* info)
   {
      countChn++;
      MLPutFunction (stdlink, "List", 5);
      /* printf ("%s %i %i %i %i\n", info->chName, info->dataRate, 
             info->ifoId, info->dcuId, info->dataType);*/
      MLPutString (stdlink, info->chName);
      MLPutInteger (stdlink, info->dataRate);
      MLPutInteger (stdlink, info->ifoId);
      MLPutInteger (stdlink, info->dcuId);
      MLPutInteger (stdlink, info->dataType);
      return 0;
   }


extern "C" 
   void ndsgetchannelinfo ()
   {
      int len = gdsChannelListLen (-1, 0);
      MLPutFunction (stdlink, "List", len);
      countChn = 0;
      gdsChannelListLen (-1, addChannelInfo);
   }


/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Interface: NDSGetData						*/
/*                                                         		*/
/*----------------------------------------------------------------------*/


extern "C" 
   void writeData (int part, int index, 
                  const float x[], int len, ndschninfo* user)
   {
      if ((part != 0) || (user == 0) || (user->x == 0)) {
         return;
      }
      if (index + len > user->ndata) {
         len = user->ndata = index;
      }
      if (len <= 0) {
         return;
      }
      memcpy (user->x + user->complex * index, x, 
             len * user->complex * sizeof (float));
   }


template <class T>
   inline void convertData (float x[], T y[], int len)
   {
      for (int i = 0; i < len; i++) {
         x[i] = static_cast<float>(y[i]);
      }
   }


extern "C"
   void ndsgetdatagps (const char* channels, double rate,
                     double fdown, double gpssec, double gpsnsec,
                     double duration)
   {
      // check parameters
      if (channels == 0) {
         MLPutInteger (stdlink, -1);   
         return;
      }
      if ((gpssec < 0) || (gpsnsec < 0)) {
         MLPutInteger (stdlink, -2);   
         return;
      }
      if (rate <= 0) {
         MLPutInteger (stdlink, -3);   
         return;
      }
      if (fdown < 0) {
         MLPutInteger (stdlink, -4);   
         return;
      }
      if (duration <= 0) {
         MLPutInteger (stdlink, -5);   
         return;
      }
      // setup NDS
      if (!connectNDS()) {
         MLPutInteger (stdlink, -6);   
         return;
      }
      if (!addChannelsNDS (channels)) {
         MLPutInteger (stdlink, -7);   
         nds.close();
         return;
      }
      if (clist.size() == 0) {
         MLPutInteger (stdlink, -8);   
         nds.close();
         return;
      }
   
      // caluclate exact start/stop time and # of points
      tainsec_t start = (tainsec_t)(gpssec * __ONESEC) + 
         (tainsec_t)gpsnsec;
      tainsec_t dur = (tainsec_t) (duration * 1E9 + 0.5);
      int ndata;
      if (rate >= 1) {
         int fSample = (int) (rate + 0.5);
         int ndec;
         if (frexp((double) fSample, &ndec) != 0.5) {
            MLPutInteger (stdlink, -9);   
            clist.clear();
            nds.close();
            return;
         }
         tainsec_t dt = _ONESEC / fSample;
         start = _ONESEC * (start / _ONESEC) +
            ((start % _ONESEC + dt/2) / dt) * dt;
         ndata = dur / dt;
         dur = ndata * dt;
         rate = (double)fSample;
      }
      else {
         int tSample = (int) (1./rate + 0.5);
         int ndec;
         if (frexp((double) tSample, &ndec) != 0.5) {
            MLPutInteger (stdlink, -9);   
            clist.clear();
            nds.close();
            return;
         }
         start = ((start + _ONESEC/2) / _ONESEC) * _ONESEC;
         ndata = dur / (tSample * _ONESEC);
         dur = ndata * (tSample * _ONESEC);
         rate = 1./(double)tSample;
      }
   
      // allocate return array
      int complex = (fdown > 0) ? 2 : 1;
      long dim[2];
      dim[0] = clist.size() + 1;
      dim[1] = ndata * complex;
      float* x = new (nothrow) float [dim[0] * dim[1]];
      if (x == 0) {
         MLPutInteger (stdlink, -10);   
         clist.clear();
         nds.close();
         return;
      }
      memset (x, 0, dim[0] * dim[1] * sizeof (float));
   
      // initialize preprocessor
      tainsec_t 	maxdelay = 0;
      int 		index = 0;
      for (ndschnlist::iterator iter = clist.begin();
          iter != clist.end(); iter++, index++) {
         iter->fPre = initPreproc (start, dur, 
                              iter->fInfo.dataRate, rate, fdown, 
                              dur, dur,
                              (PartitionDataOut)writeData, &(*iter));
         iter->ndata = ndata;
         iter->complex = complex;
         iter->x = x + ndata * (index + 1) * complex;
         if (iter->fPre == 0) {
            MLPutInteger (stdlink, -11);   
            delete [] x;
            clist.clear();
            nds.close();
            return;
         }
         if (iter->fPre->delayshift > maxdelay) {
            maxdelay = iter->fPre->delayshift;
         }
      }
   
      // start NDS
      taisec_t ndsstart = (start - (maxdelay + maxdelay/10)) / _ONESEC;
      taisec_t ndsduration = (dur + (maxdelay + maxdelay/5)) / _ONESEC;
      while ((ndsstart + ndsduration) * _ONESEC < 
            start + dur + (maxdelay + maxdelay/10)) {
         ndsduration++;
      }
      if (nds.RequestData (ndsstart, ndsduration) != 0) {
         MLPutInteger (stdlink, -12);
         delete [] x;
         clist.clear();
         nds.close();
         return;
      }
   
      // wait for data
      const int size = 16*1024;
      float		fdat[size];	// data buffer
      char* buf = 0;
      tainsec_t		keepalivetime;	// keep allive time
      keepalivetime = TAInow();
      while (nds.GetData (&buf) > 0) {
         DAQDRecHdr*	head = (DAQDRecHdr*) buf;
         char*		dptr = buf + sizeof (DAQDRecHdr);
         int		datasize =	// size of data
            head->Blen - (sizeof (DAQDRecHdr) - sizeof (int));
         int		idata = 0;	// data index
         tainsec_t	time;
         time = (tainsec_t)head->GPS * _ONESEC + head->NSec;
      
      #ifdef DEBUG
         cout << "recv data @ " << head->GPS << ":" << head->NSec <<
            " for " << head->Secs << " (len = " << head->Blen - 16 << 
            ", seq = " << head->SeqNum << ")" << endl;
      #endif
         // sprintf (msg, "Print [\"receive data at GPS %i\"]", head->GPS);
         // MLEvaluateString (stdlink, msg);
      
         // keep alive
         if (keepalivetime + _KEEPALIVEINTERVAL*_ONESEC < TAInow()) {
            MLEvaluateString (stdlink, (char*)"Null");
            keepalivetime = TAInow();
         }
      
         // find channel and call preprocessor
         for (DAQSocket::Channel_iter iter = nds.mChannel.begin();
             iter != nds.mChannel.end(); iter++) {
            // calculate # of data points
            int ndata = iter->second.mRate * head->Secs;
            // check buffer length
            if (idata + ndata * iter->second.mBPS > datasize) {
               MLPutInteger (stdlink, -13);
               delete [] x;
               clist.clear();
               nds.close();
               return;
            }
            // find daq channel and invoke callback
            unsigned int index1 = 0;
            for (ndschnlist::iterator it = clist.begin(); 
                it != clist.end(); it++, index1++) {
               if (gds_strcasecmp (it->fInfo.chName, 
                                  iter->first.c_str()) == 0) {
                  break;
               }
            }
            if (index1 >= clist.size()) {
               idata += ndata * iter->second.mBPS;
               dptr += ndata * iter->second.mBPS;
               continue;
            }
            float* fptr = 0;
            // check if data has to be converted into floats
            if (clist[index1].fInfo.dataType == DAQ_DATATYPE_FLOAT) {
               fptr = (float*) dptr;
            }
            else {
            // allocate buffer
               if (ndata <= size) {
                  fptr = fdat;
               }
               else {
                  fptr = new (nothrow) float [ndata];
                  if (fptr == 0) {
                     idata += ndata * iter->second.mBPS;
                     dptr += ndata * iter->second.mBPS;
                     continue;
                  }
               }
               switch (clist[index1].fInfo.dataType) {
                  case DAQ_DATATYPE_16BIT_INT: 
                     {
                        convertData (fptr, (short*) dptr, ndata);
                        break;
                     }
                  case DAQ_DATATYPE_32BIT_INT:
                     {
                        convertData (fptr, (int*) dptr, ndata);
                        break;
                     }
                  case DAQ_DATATYPE_64BIT_INT:
                     {
                        convertData (fptr, (long long*) dptr, ndata);
                        break;
                     }
                  case DAQ_DATATYPE_DOUBLE:
                     {
                        convertData (fptr, (int*) dptr, ndata);
                        break;
                     }
                  default:
                     {
                        memset (fptr, 0, ndata * sizeof (float));
                        break;
                     }
               }
            }
         
            // invoke proprocessor
            preprocess (clist[index1].fPre, time, fptr, ndata);
            // free data buffer if necessary
            if ((ndata > size) && 
               (clist[index1].fInfo.dataType != DAQ_DATATYPE_FLOAT)) {
               delete [] fptr;
            }
            // advance to next channel
            idata += ndata * iter->second.mBPS;
            dptr += ndata * iter->second.mBPS;
         }
         delete [] buf;
         buf = 0;
      }
      delete [] buf;
   
      // write time
      for (int i = 0; i < ndata; i++) {
         x[complex * i] = (double) i / rate;
      }
   
      // send data back: check for abort first
      if (MLAbort) {
         MLPutInteger (stdlink, -99);   
      }
      else {
         MLPutFloatArray (stdlink, x, dim, NULL, 2);
      }
      delete [] x;
      clist.clear();
      nds.close();
   }


extern "C" 
   void ndsgetdatautc (const char* channels, double rate,
                     double fdown, int year, int month, int day,
                     int hour, int min, double sec, double duration)
   {
      double 		gpssec;
      double		gpsnsec;
      utc_t		utc;
      utc.tm_year = year - 1900;
      utc.tm_mon = month - 1;
      utc.tm_mday = day;
      utc.tm_hour = hour;
      utc.tm_min = min;
      utc.tm_sec = (int) sec;
      gpssec = (double) UTCtoTAI (&utc);
      gpsnsec = (sec - utc.tm_sec) * 1E9;
   
      ndsgetdatagps (channels, rate, fdown, gpssec, gpsnsec, duration);
   }



/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Interface: NDSGetTrend						*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

extern "C"
   void ndsgettrendgps (const char* channels, double rate,
                     double gpssec, double gpsnsec,
                     double duration, const char* format)
   {
      // check parameters
      if (channels == 0) {
         MLPutInteger (stdlink, -1);   
         return;
      }
      if (gpssec < 0) {
         MLPutInteger (stdlink, -2);   
         return;
      }
      if (rate <= 0) {
         MLPutInteger (stdlink, -3);   
         return;
      }
      if (duration <= 0) {
         MLPutInteger (stdlink, -4);   
         return;
      }
      if (!format || 
          ((strcmp (format, "y") != 0) && 
	   (strcmp (format, "nysmM") != 0))) {
         MLPutInteger (stdlink, -10);   
         return;
      }
      bool simpletrend = (strcmp (format, "y") == 0);
      // setup NDS
      if (!connectNDS()) {
         MLPutInteger (stdlink, -5);   
         return;
      }
      if (!addChannelsNDS (channels, true, simpletrend)) {
         MLPutInteger (stdlink, -6);   
         nds.close();
         return;
      }
      if (clist.size() == 0) {
         MLPutInteger (stdlink, -7);   
         nds.close();
         return;
      }
   
      // determine decimation factor
      bool mintrend;
      int dec;
      int dt;
      rate = fabs (rate);
      if (rate <= 1./60.+1E-12) {
         mintrend = true;
         dec = (int) (1./60. / rate + 0.5);
         if (dec < 1) dec = 1;
         dt = 60 * dec;
      }
      else {
         mintrend = false;
         dec = (int) (1. / rate + 0.5);
         if (dec < 1) dec = 1;
         dt = dec;
      }
   
      // determine number of data points
      int ndata = (int) duration / dt;
      int depth = simpletrend ? 2 : 6;
   
      // allocate memory for return
      double* data = new (nothrow) double [clist.size() * ndata * depth];
      if (data == 0) {
         MLPutInteger (stdlink, -8);
         clist.clear();
         nds.close();
         return;
      }
      memset (data, 0, clist.size() * ndata * depth * sizeof (double));
   
      // start NDS trend reader
      unsigned long start = (int) (gpssec + 0.5);
      if (mintrend) {
         //utc_t	utc;
         //TAItoUTC (start + 30, &utc);
         //utc.tm_sec = 0;
         //start = UTCtoTAI (&utc);
         start = 60 * ((start + 30) / 60);
      }
      if (nds.RequestTrend (start, ndata * dt, mintrend) != 0) {
         MLPutInteger (stdlink, -9);
         delete [] data;
         clist.clear();
         nds.close();
         return;
      }
   
      // wait for data
      char* buf = 0;
      while (nds.GetData (&buf) > 0) {
         DAQDRecHdr*	head = (DAQDRecHdr*) buf;
         char*		dptr = buf + sizeof (DAQDRecHdr);
      
      #ifdef DEBUG
         cout << "recv trend @ " << head->GPS << ":" << head->NSec <<
            " for " << head->Secs << " (len = " << head->Blen - 16 << 
            ", seq = " << head->SeqNum << ")" << endl;
      #endif
      
         // copy data into return array directly
         unsigned int index1;
         int index2;
#ifdef USE_N
         int* n;
#endif
         float* min;
         float* max;
         double* mean;
         double* rms;
         int dttrend = (mintrend ? 60 : 1);
#ifdef USE_N
	 const int adv = 5;
#else
	 const int adv = 4;
#endif
         for (DAQSocket::Channel_iter iter = nds.mChannel.begin();
             iter != nds.mChannel.end(); advance (iter, adv)) {
            // compute data pointers
	    if (!simpletrend) {
              max = (float*) dptr;
              dptr += sizeof (float) * head->Secs / dttrend;
	    }
            mean = (double*) dptr;
            dptr += sizeof (double) * head->Secs / dttrend;
	    if (!simpletrend) {
               min = (float*) dptr;
               dptr += sizeof (float) * head->Secs / dttrend;
#ifdef USE_N
               n = (int*) dptr;
               dptr += sizeof (int) * head->Secs / dttrend;
#endif
               rms = (double*) dptr;
               dptr += sizeof (double) * head->Secs / dttrend;
	    }
            // get 1st index
            string::size_type pos = iter->first.find_last_of (".");
            if (pos == string::npos) {
               break;
            }
            string name = iter->first.substr (0, pos);
            index1 = 0;
            for (ndschnlist::iterator it = clist.begin(); 
                it != clist.end(); it++, index1++) {
               if (gds_strcasecmp (it->fInfo.chName, name.c_str()) == 0) {
                  break;
               }
            }
            if (index1 >= clist.size()) {
               break;
            }
            // loop through result list
            for (int i = 0; i < head->Secs / dttrend; i++) {
               // calulate index into result
               index2 = (head->GPS - start + i * dttrend) / dt;
               if ((index2 < 0) || (index2 >= ndata)) {
                  continue;
               }
               double* dest = data + index1 * (ndata * depth) + index2 * depth;
               // check if copy or add/min/max formulae
   	       if (simpletrend) {
        	  dest[1] += mean[i]/(double)dec;
	       }
	       else {
        	  if ((dec == 1) || (dest[1] == 0)) {
                     dest[4] = min[i];
                     dest[5] = max[i];
        	  }
        	  else {
                     dest[4] = std::min ((double)min[i], dest[4]);
                     dest[5] = std::max ((double)max[i], dest[5]);
        	  }
#ifdef USE_N
        	  dest[1] += n[i];
        	  dest[2] += mean[i]*n[i];
        	  dest[3] += rms[i]*rms[i]*n[i];
#else
        	  dest[1] += 1;
        	  dest[2] += mean[i];
        	  dest[3] += rms[i]*rms[i];
#endif
               }
            } 
         }
         delete [] buf;
         buf = 0;
      }
      delete [] buf;
   
      // check for abort
      if (MLAbort) {
         MLPutInteger (stdlink, -99);   
         delete [] data;
         clist.clear();
         nds.close();
         return;
      }
   
      // normalize result and caluclate std dev instead of rms
      for (int j = 0; j < (int)clist.size(); j++) {
         for (int i = 0; i < ndata; i++) {
            double* dest = data + j * (ndata * depth) + i * depth;
            dest[0] = dt * i;
	    if (!simpletrend) {
               if (dest[1] > 0) {
        	  dest[2] = dest[2] / dest[1];
               }
               if (dest[1] > 1.5) {
        	  dest[3] = sqrt (fabs(dest[1]/(dest[1]-1.) * 
                                      (dest[3]/dest[1] - dest[2]*dest[2])));
               }
               else {
        	  dest[3] = 0;
               }
            }
         }
      }
   
      // send result back
      long dim[] = {clist.size(), ndata, depth};
      MLPutRealArray (stdlink, data, dim, NULL, 3);
      delete [] data;
      clist.clear();
      nds.close();
   }


extern "C" 
   void ndsgettrendutc (const char* channels, double rate,
                     int year, int month, int day,
                     int hour, int min, double sec, 
		     double duration, const char* format)
   {
      double 		gpssec;
      double		gpsnsec;
      utc_t		utc;
      utc.tm_year = year - 1900;
      utc.tm_mon = month - 1;
      utc.tm_mday = day;
      utc.tm_hour = hour;
      utc.tm_min = min;
      utc.tm_sec = (int) sec;
      gpssec = (double) UTCtoTAI (&utc);
      gpsnsec = (sec - utc.tm_sec) * 1E9;
   
      /*printf ("t channels = %s\nrate = %g year=%i "
             "month=%i day=%i hour=%i min=%i, sec=%g duration=%g\n", 
             channels, rate, year, month, day, hour, min, sec, duration);*/
      ndsgettrendgps (channels, rate, gpssec, gpsnsec, duration, format);
   }

/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Main program								*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#if MACINTOSH_MATHLINK

   int main( int argc, char* argv[])
   {
        /* Due to a bug in some standard C libraries that have shipped with
         * MPW, zero is passed to MLMain below.  (If you build this program
         * as an MPW tool, you can change the zero to argc.)
         */
      argc = argc; /* suppress warning */
      return MLMain( 0, argv);
   }

#elif WINDOWS_MATHLINK

   int PASCAL WinMain( HANDLE hinstCurrent, HANDLE hinstPrevious, 
                     LPSTR lpszCmdLine, int nCmdShow)
   {
      char  buff[512];
      char FAR * argv[32];
      char FAR * FAR * argv_end = argv + 32;
   
      if( !MLInitializeIcon( hinstCurrent, nCmdShow)) 
         return 1;
      MLScanString( argv, &argv_end, &commandline, &buf);
      return MLMain( argv_end - argv, argv);
   }

#else

   int main (int argc, char* argv[])
   {
      return MLMain (argc, argv);
   }

#endif
