//
//    Bit test function
//
#ifndef LISTEN_HH
#define LISTEN_HH

#include "DatEnv.hh"
#include "Pipe.hh"
#include "laudio.hh"
#include <string>

class AudioPanel;


/**  @name Listen
  *  The Listen monitor plays a data channel in real-time over the 
  *  speakers on the default audio device of a DMT workstation.
  *  At present a single channel is played. The channel data are 
  *  resampled to 8kHz by linear interpolation to match the usual
  *  sampling rate for audio output devices. Optional high pass 
  *  and low-pass filters may be applied to the data before the 
  *  data are sent to the audio device.
  *
  *  {\large {\bf Running Listen } }
  *
  *  The Listen command has the following syntax:
  *
  *  \begin{verbatim}
     Listen [<dmt-args>] [-gain <gain>] [-high <fHi>] [-low <fLow>] <channel>
     \end{verbatim}
  *
  *  Where: \\
  *
  *  \begin{tabular}{|l|l|}
  *  #<dmt-args># & The standard DMT DatEnv arguments (e.g. -infile) \\
  *  #-gain <gain># & specify an absolute gain. If the gain isn't specified
  *                 Listen will guess at a gain that will make the signal 
  *                 audible. \\
  *  #-high <fHi># & specify a maximum frequency. A filter is added 
  *                with two real poles at #<fHi># and two zeros at fNyquist. \\
  *  #-low <fLow># & specify a minimum frequency. A filter is added with two 
  *                real poles at #<fLow># and two zeros at 0 Hz. \\
  *  #<channel># & is the name of the channel to be listened to.
  *  \end{tabular}
  *
  *  When Listen is launched, a control panel is displayed on the device 
  *  specified by the #DISPLAY# environment variable. By default all audio 
  *  ports are disabled. You must select the appropriate port.
  *
  *  @memo Listen to a data channel
  *  @author John Zweizig.
  *  @version 1.2; Last Modified July 25, 2002
  */
//{
//}


class Listen : public DatEnv {
  public:
    Listen(int argc, const char *argv[]);
    ~Listen();
    void ProcessData(void);
    void Attention(void);

  private:
    std::string mChannel;
    LAudio mPort;
    AudioPanel* mPanel;
    double mVolume;
    int    mMode;
    bool   mMute;
    double mGain;
    std::string mFDefine;
    double mHigh;
    double mLow;
    Pipe*  mFilter;
};

#endif  // !defined(LISTEN_HH)
