#ifndef TimingCollector_HH
#define TimingCollector_HH

#include <sys/types.h>
#include <sys/socket.h>
#include <pthread.h>
#include <string>
#include <inttypes.h>
#include "gmutex.hh"
#include "Time.hh"
#include "TimingTypes.hh"

/** The TimingCollector program reads the state of the timing system.
  *
  * command line arguments:
  * -f <filename> : name of configuration file
  * -a <ip> : ip address of RS422 converter
  * -p <port> : port number of RS422 converter (default is 1234)
  * -m <smem> : name of shared memory partition (default is TimingSystem)
  * -s <len> : size of buffers in bytes (default is 1 MB)
  * -n <num> : number of buffers in partition (default is 8)
  * -l 'file': name of log file
  * -1 : run once and dump dat
  * -d : print debug messages
 * @brief Timing Collector program
  *
  */


// Default RS422 converter port
int kPORTDEFNUM = 1235;

// magic string
const char* const kMagicString = "LIGO TIMING SYSTEM VERSION 1.0\r\n";
// transmitted data length (without magic string but with CRC value
const int kDataLenV1 = 16*(1024/8) + 2048/8 + 4;

/// Port collector
class port_collector {
public:
   /// Constructor
   port_collector (const char* portname, int portnum = -1, 
                   const char* name = 0) 
   : fOpened (false), fError (0), fPortname (portname ? portname : ""),
     fTID (0), fName (name ? name : ""), fTime (0,0), 
     fSocket (-1), fPort (portnum) {}

   /// Read from the port
   void* read_port ();
   
   /// Port open
   bool 		fOpened;
   /// Error condition
   int			fError;
   /// Port name
   std::string 		fPortname;
   /// Thread id
   pthread_t 		fTID;
   /// Mutex
   mutable thread::mutex fMux;
   /// Name of module
   std::string		fName;

   /// Time stamp of data
   Time			fTime;
   /// Data
   mfo_t		fData;

protected:
   /// Socket
   int  		fSocket;
   /// Port number
   int			fPort;
};

#endif /* TimingCollector */
