#include <iostream>
#include <fstream>
#include "getStuff.hh"
#ifndef __CINT__
#include "ParseLine.hh"
#include "Dacc.hh"

#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/bigStuff/getStuff.cc,v 1.3 2004/07/12 05:48:41 jzweizig Exp $"
#define PIDTITLE "bigStuff_get"
#include "ProcIdent.hh"
#endif   //__CINT__

EXECDAT(getStuff);

using namespace std;

//======================================  Skeleton monitor constructor
getStuff::getStuff(int argc,const char *argv[])
  : DatEnv(argc,argv),mCount(0),mPrevious(0),mConfig(""), mOutput(""),
    mOSC(getDacc()) {
  if (Debug() > 2) cout << "Starting getStuff..." << endl;
  const char *cfile="getStuff.conf";
  const char *oscfile="bigStuff_osc.conf";
  const char *offfile=0;
  const double stride(1.0);

  ///////////////////////  Read the configuration file
  // Note to self:  Don't use finish() in the Constructor as it must complete
  // constructor before exiting.
  for (int i=1; i<argc; i++) {
    if (!strcmp("-conf", argv[i])) {
      if (chkNextArg(i,argv,argc)) cfile = argv[++i];
      else {
	cerr << "Couldn't find configuration file"<<endl;
	exit(Usage(argv[0])+1);
      }
    } else if (!strcmp("-osc", argv[i])) {
      if (chkNextArg(i,argv,argc)) oscfile = argv[++i];
      else {
	cerr << "Couldn't find OSC file"<<endl;
	exit(Usage(argv[0])+1);
      }
    } else if (!strcmp("-offline", argv[i])) {
      if (chkNextArg(i,argv,argc)) offfile = argv[++i];
      else {
	cerr << "Couldn't find file containing offline directories."<<endl;
	exit(Usage(argv[0])+1);
      }
    } else if (!strcmp("-verbose", argv[i])) {
      if (chkNextArg(i,argv,argc)) exit(Usage(argv[0]));
      else {verbose=true; continue;}
    } else if (!strcmp("-h",argv[i])|| !strcmp("--help",argv[i])) {
      Usage(argv[0]);
      Help();
      exit(0);
    } else if (isDatEnvArg(argv[i])) {i++;
    } else exit(Usage(argv[0]));
  }
  //=====================  Define and initialize parameters
  mDict.addPar("Block",     int(60));
  mDict.addPar("RunTime",    int(1440));
  mDict.addPar("OutputFile", string(""));
//    mDict.addPar("OSCFile",    string(""));

  //=====================  Check for offline files
  if (offfile) {
    std::ifstream offstr(offfile);
    if (!offstr) {
      cerr << "Couldn't open file containing offline directories."<<endl;
      exit(Usage(argv[0])+2);
    }
    char mLine[1024];
    while (!offstr.eof()) {
      offstr.getline(mLine, sizeof(mLine));
      getDacc().addFile(mLine);
    }
  }

  //=====================  Read in the configuration files
  mConfig = cfile;
  mOSC.readConfig(oscfile);
  ReadConfig();

  //=====================  Connect the channels to Dacc
  pTS = new TSeries*[mChans.size()];
  for (unsigned int i = 0; i < mChans.size(); i++) {
    pTS[i] = 0;
    if (! getDacc().isChannelRead(mChans[i].name.c_str())) {
      getDacc().addChannel(mChans[i].name.c_str(),0,&pTS[i]);
      if (Debug()) cout <<"Adding channel "<<mChans[i].name.c_str()<<endl;
    } 
  }

  //=====================  Set up the rest of the variables
  getDacc().setStride((float) stride);
  mBlock = mDict.getInt("Block");
  mBlockCount = 0;
  mTimer = mDict.getInt("RunTime");
  mOutput = mDict.getString("OutputFile");
  getDacc().synch();
  if (Debug()) cout <<"Filled first set of data."<<endl;
  mPrevious = getDacc().getTime() - getDacc().getStride();
//    if (!mDict.getString("OSCFile").empty()) {
//      mOSC.readConfig(mDict.getString("OSCFile").c_str());
//    }
}

//====================================  Read the config file
void getStuff::ReadConfig(void) {
  if (Debug() > 2) cout<<"Reached ReadConfig(void)"<<endl;
  //=====================  Set up the line parser
  ParseLine lp(mConfig.c_str());
  if (Debug()) lp.setLog(cout);
  int nLine(0);
  for (int nArg=lp.getLine(); nArg>=0; nArg=lp.getLine()) {
    nLine++;
    //---------  Empty command line
    if (!lp[0]) {
      continue;
    //---------  Parse a 'Parameter' line
    } else if (!strcmp(lp[0],"Parameter")) {
      if (nArg != 3) {
	cerr<<"Parameter not specified, line:"<<nLine<<endl;
	continue;
      } else for (int i=1; i<nArg; i+=2) {
	if (!mDict.exists(lp[i])) {
	  cerr<<"No such parameter: "<<lp[i]<<endl;
	  break;
	} else {
	  if (Debug()) cout<<"Set parameter: "<<lp[i]<<endl;
	  mDict.setPar(lp[i], lp[i+1]);
	}
      }
    //---------  Add a channel for monitoring
    } else if (nArg == 1) {
      if (Debug()) cout<<"Adding channel: "<<lp[0]<<endl;
      mChans.push_back(getChannelStat(lp[0]));
    //---------  Invalid statement
    } else {
      cerr<<"Invalid command in "<<mConfig<<", line "<<nLine<<": "
	  <<lp[0]<<endl;
    }
  }
}

//======================================  Skeleton monitor destructor
getStuff::~getStuff(void) {
  ofstream ofile(mOutput.c_str(), ios::out | ios::app);
  //=====================  Test whether we can write to file; if not, use cout
  if (Debug() > 1) 
    cout <<"Test this statement-- "<<mOutput<<" is open: "
	 <<ofile.is_open()<<endl;
  ostream *outputStr;
  ofile.is_open() ? outputStr = &ofile : outputStr = &cerr;
  //=====================  Write out any missing strides
  *outputStr<<"#Final stats-- lost data:"<<endl;
  for (vector<pair<Time,Interval> >::const_iterator j = mSkips.begin();
       j != mSkips.end(); j++) {
    *outputStr<<"#Data skipped at "<<j->first.getS()<<" for "
	      <<double(j->second)<<" seconds."<<endl;
  }
  //=====================  Write out the channel statistics
  *outputStr<<"#Final stats-- Channel statistics: <name> <mean> <sdev> "
	    <<"<avg dev> <max> <min>"<<endl;
  outputStr->precision(6);
  for (unsigned long int  i(0); i < mChans.size(); i++) {
    //---------  Comment out any data that's not changing (much)
    if (abs(mChans[i].var) < .0001) *outputStr<<"#";
    //---------  Comment out any data that's changing too much
    if ((mChans[i].max - mChans[i].min) > 65500) *outputStr<<"#";
    *outputStr<<mChans[i].name<<"\t"<<mChans[i].mean<<"\t"<<mChans[i].mean2
	      <<"\t"<<sqrt(mChans[i].var)<<"\t"<<mChans[i].adev<<"\t"
	      <<mChans[i].max<<"\t"<<mChans[i].min<<endl;
    //---------  might as well delete the data array, now that we're here
    delete[] mData[i];
  }
  //=====================  Unload any data reserved by pointers.
  delete[] pTS;
  if (Debug() > 2) cout << "getStuff is finished." << endl;
}

//======================================  The data crunching routine
void getStuff::ProcessData(void) {
  if (mOSC.satisfied("Mode_Cleaner_locked")) {
    if (Debug() > 2)  cout << "Processing Data" << endl;
    if (mCount == 0) {
      for (unsigned int i(0); i < mChans.size(); i++) {
	mChans[i].numPts = pTS[i]->getNSample();
//        mChans[i].numPts = 256;
	mData.push_back(new double[mChans[i].numPts*mBlock]);
	if (Debug()) 
	  cout <<"Adding "<<mChans[i].numPts*mBlock
	       <<"-pt double array to mData."<<endl;
      }
    }
    //-----------  Set the timestamps
    mCurrent = pTS[0]->getStartTime();
    //=====================  Check if there's a gap in data; add to mSkips
    if (! Almost((mPrevious + pTS[0]->getInterval()),mCurrent) || 
	!pTS[0]->isGood()) {
      mSkips.push_back(pair<Time,Interval>
		       (mPrevious,Interval(mCurrent -  
					   (mPrevious + 
					    pTS[0]->getInterval()) ) ));
    }
    //=====================  Calculate the stats for the current set of frames
    double curMean2,curVar,curADev;
    vector<double> curMean(mChans.size());
    long int curMax(0),curMin(0);
    
    for (unsigned int i=0; i < mChans.size(); i++) {
      //---------  Gather the TSeries information
      curMax = (long int)(pTS[i]->getMaximum());
      curMin = (long int)(pTS[i]->getMinimum());
      curMean[i] += pTS[i]->getAverage();
      //---------  Update the channel max,mean
      if (mChans[i].max < curMax) mChans[i].max = curMax;
      if (mChans[i].min > curMin) mChans[i].min = curMin;
      //---------  Get the channel data
      pTS[i]->getData(mChans[i].numPts,
		      mData[i] + mChans[i].numPts * mBlockCount);
      cout << "mBlock = "<<mBlock<<", mBlockCount = "<<mBlockCount
	   << ", start address of new array = "
	   <<mData[i] + mChans[i].numPts * (mBlockCount)<<endl;
    }
    mBlockCount++;
    //=====================  When we've filled "Block" seconds'-worth of data
    if (mBlockCount == mBlock) {
      cout << "Entering calculation mode: mBlockCount = "<<mBlockCount<<endl;
      mBlockCount = 0;
      for (unsigned int i=0; i < mChans.size(); i++) {
	curMean2 = curVar = curADev = 0.0;
	//-------  Calculate the current mean (up to now it's Sum(<x>))
	curMean[i] /= mChans.size();
	//------- Meanings for the following variables: tmpSx= Sum(x-<x>),
	//------- tmpSx2 = Sum((x - <x>)^2), tmpSax = Sum(abs(x - <x>))
	double tmpavg(0.0),tmpSx(0.0),tmpSx2(0.0),tmpSax(0.0);
	for (long int j(0); j < mChans[i].numPts * mBlock; j++) {
	  //-----  recalculate the current mean, verify data copy is correct
	  tmpavg += mData[i][j];
	  //-----  Intermediate step: Calculate (x_i - <x>)
	  tmpSx += mData[i][j] - curMean[i];
	  tmpSx2 += (mData[i][j] - curMean[i]) * (mData[i][j] - curMean[i]);
	  tmpSax += abs(mData[i][j] - curMean[i]);
	}
	//-------  Alternative mean
	curMean2 = tmpavg / (mChans[i].numPts * mBlock);
	//-------  Current Std Dev = sqrt(Sum((x_i - <x>)^2) / (N-1))
	//-------           which is close to <x^2> - <x>^2 - N*(<x - <x>>)^2
	curVar = tmpSx2 / (mChans[i].numPts * mBlock - 1);
	// or curSDev = sqrt( (tmpSx2 - tmpSx * tmpSx / mChans[i].numPts) / 
	//                    (mChans[i].numPts - 1) );  
	//                ^--- This is the corrected two-pass algorithm
	//-------  Current Avg Dev = Sum(abs(x_i - <x>))/N
	curADev = tmpSax / (mChans[i].numPts * mBlock);
	cout <<"Temporary stats: "<<curMean[i]<<" "<<curMean2<<" "
	     <<sqrt(curVar)<<" "<<curADev<<endl;
	//=================  Calculate the stats up to this point
	mChans[i].mean = (curMean[i] + mCount * mChans[i].mean) / (mCount + 1);
	mChans[i].mean2 = (curMean2 + mCount * mChans[i].mean2) / (mCount + 1);
	double num1(mChans[i].numPts * mBlock - 1), 
	  num2(mCount * mChans[i].numPts * mBlock - 1),
	  denom((mCount + 1) * mChans[i].numPts * mBlock - 1);
	mChans[i].var = (num1 * curVar + num2 * mChans[i].var) / denom;
	mChans[i].adev = (curADev + mCount * mChans[i].adev) / (mCount + 1);
	if (Debug() > 1) {
	  cout.precision(6);
	  cout << "Stats for "<<mChans[i].name<< " at "<<mCurrent.totalS()
	       <<": "<<mChans[i].mean<<" "<<mChans[i].mean2<<" "
	       <<sqrt(mChans[i].var)<<" "<<mChans[i].adev<<" "
	       <<mChans[i].max<<" "<<mChans[i].min<<endl;
	}
      }
      if (Debug()) 
	cout <<"Got through 'for (unsigned int i(0); i<mChans.size(); i++)'"
	     <<"loop..."<<endl;
    }
    mPrevious = mCurrent;
    if (++mCount == (mTimer * mBlock)) {
      if (Debug() > 2) cout <<"Reached my goal; finishing..."<<endl; 
      finish();
    }
    if (Debug())
      cout <<"Finishing Process Data... mCount= "<<mCount<<", mTimer*mBlock= "
	   <<(mTimer * mBlock)<<endl;
  }
}

bool getStuff::chkNextArg(int i, const char* argv[], const int argc) {
  if (++i < argc) {
    std::string args(argv[i]);
    if (args[0] != '-') return true;
  } // Note: Doesn't check next arg for number value
  return false;
}

int getStuff::Usage(const char* progName) {
  cerr << "Usage: "<< progName<<" [-conf <file>] [-osc <file>] "
       << "[-offline <file>] [-verbose] [-h | --help]"<< endl;
  return 1;
}

void getStuff::Help(void) {
  cerr <<"getStuff gets statistics for a number of channels and writes them to a file,\n"
       <<"which is then used by findStuff to look for large deviations in the data time\n"
       <<"series."<<endl;
}
