#include <TidClass.h>
#include <errors.h>
#include <string.h>
#include <math.h>

#ifdef WithDMT
#include <TrigBase.hh>
#include <TrigClient.hh>
#include <TidClassBackground.h>
extern TidClassBackground * gTidClass;
#endif

void TidEventsDef::SetmaxAboveThrStop (unsigned int m) {

  if(rect) rect->SetmaxAboveThrStop(m);

}


TidClass::TidClass(int argc, const char *argv[]) {

  UpTime = 0;
  UpTimeRef = -1U;
  UpTimeMax = 0;

  DoTrigger = 1;

  string confFile("./.TidClass");

  DoDumpCat = 0;

  if(argc==2) confFile = string(argv[1]);


  try {
    TidMessage("Setting up socket connection to data server...");
    InitIO(confFile);
    
    string tmpstr("Connected to: ");
    list<string>::iterator so;
    for(so=source_.begin(); so!=source_.end(); so++)
      tmpstr+=*so + string(" ");
    TidMessage(tmpstr);

    TidMessage("Reading event catalog...");
    InitCatalog();
  }
  catch (TidError err) {HandleError(err); exit(-1);}

  if(DoDumpCat) {

    const char * fout = DumpCatFile.c_str();

    DumpCat(fout);

    cout<<"Catalog dumped to file "<<fout<<endl;

  }

  if(DoReport) {
    void *ReportLoop(void *s);

    if(pthread_create(&RLooper, NULL, ReportLoop, this))
      {cerr<<"Thread creation error\n"; 
      exit(-1);}
  }
}



void *ReportLoop(void *s) {
  /*
  TidClass *t = (TidClass *)s;

  for(;;) {
    try {
      t->Report();
    }
    catch (TidError err) {HandleError(err);}

    usleep(1000000);
  }
  */
  return NULL;
}


  

TidClass::~TidClass () {

  pthread_cancel(RLooper);

  list<TidIO*>::iterator p;
  for(p=tio.begin();p!=tio.end();p++)
    delete *p;
  
  #ifdef WithDMT
  if(HtmlDef) {
    delete htmlTable;
  }
  #endif

  if(ASCIIDef) {
    ASCIIStream->close();
    delete ASCIIStream;
  }
}


void TidClass::InitIO(string confFile) {

  FILE* in;
  char buf[2048];
  unsigned int lnumb = 0;
  list<string> server;
  list<int> port;

  if(!(in=fopen(confFile.c_str(),"r"))) 
    throw TidError(string("Unable to open configuration file ") + confFile + string(" for reading."));

  bool ServerDef = 0,
    PortDef = 0,
    GraceTimeDef = 0,
    CatalogDef = 0;

  HtmlDef = 0;
  DoReport = 0;
  
  for(;;) {

    fgets(buf, 2048, in);

    lnumb++;

    if(buf[0]!='#' && buf[0]!='\n') { // check for comments or empty lines
      
      char *ptr;
      bool knownOption = 0;

      // General Options

      if(strstr(buf,"Server") == buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	char* ptr2 = strrchr(buf,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';
	server.push_back(string(ptr+1));
	ServerDef=1;
      }

      if(strstr(buf,"Html") == buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	char* ptr2 = strrchr(buf,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';
	htmlFile = string(ptr+1);
	HtmlDef=1;
      }


      if(strstr(buf,"StatusReport") == buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	char* ptr2 = strrchr(buf,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';
	ReportFile = string(ptr+1);
	DoReport=1;
      }


      if(strstr(buf,"DumpCatalog") == buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	char* ptr2 = strrchr(buf,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';
	DumpCatFile = string(ptr+1);
	DoDumpCat=1;
      }


      if(strstr(buf,"Ascii") == buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	char* ptr2 = strrchr(buf,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';
	ASCIIFile = string(ptr+1);
	ASCIIDef=1;
      }

      if(strstr(buf,"NoLDASTrigger") == buf) {
	knownOption = 1;
	DoTrigger = 0;
      }

      if(strstr(buf,"Port") == buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	port.push_back(atoi(ptr+1));
	PortDef=1;
      }

      if(strstr(buf,"Catalog")==buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	char* ptr2 = strrchr(buf,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';
	catPath = string(ptr+1);
	CatalogDef=1;
      }

      if(strstr(buf,"GraceTime") == buf) {
	knownOption = 1;
	ptr = strchr(buf,'=');
	GraceTime = atof(ptr+1);
	GraceTimeDef=1;
      }

      if(!knownOption) 
	TidWarning("unknown option in configuration file.");
      
    }

    if(feof(in)!=0) break;  

  }

  fclose(in);

  if(!ServerDef || !PortDef || !CatalogDef || !GraceTimeDef)
    throw TidError("Missing option in configuration file.");

  if(ASCIIDef) {
    ASCIIStream = new std::ofstream(ASCIIFile.c_str());

    if(!(*ASCIIStream)) 
      throw TidError ( string("Unable to open file ") + ASCIIFile );
  }

  #ifdef WithDMT
  // set up html writer
  if(HtmlDef) {
    htmlTable = new html::table("Classified Events");

    htmlTable->addColumn("Event name");
    htmlTable->addColumn("Start time");
    htmlTable->addColumn("Cost");
    htmlTable->addColumn("Variable values");
    htmlTable->setBorder(1);

    
  }
  #endif

  //CONNECTS TO PORT, OBTAIN CHANNEL INFO//
  list<string>::iterator ser = server.begin();
  list<int>::iterator po = port.begin();
  double TimeResOld;
  unsigned int id = 0;

  TimeResOld = -1.0;

  for(unsigned int i=0;ser!=server.end(); ser++, po++, i++) {

    try {
      TidIO* tmptio = new TidIO((*ser).c_str(),*po);
      tio.push_back(tmptio);
	
      string source__;
      list<string> chnames__;
      list<unsigned int> cid__, fs__;

      try
	{
	  while(!(tmptio->ReadTableofContent(&source__, &chnames__, &cid__, &fs__,&TimeResolution, &T0)));
	}

      catch (TidError err) {HandleError(err);}


      if(TimeResOld < 0.0 || TimeResolution == TimeResOld)
	{
	  TimeResOld = TimeResolution;
	  
	  source_.push_back(source__);
	  
	  list<unsigned int>::iterator p,q;
	  list<string>::iterator r;
	  for(p=cid__.begin(), q = fs__.begin(), r = chnames__.begin(); p!=cid__.end(); p++, q++, r++) 
	    {
	      fs_.push_back(*q);
	      chnames_.push_back(*r);
	      cid_.push_back((*p) + 1000 * i);
	    }
	}
	
      else 
	TidWarning("Not all data servers have same time resolutions...");
    

      chnames__.clear();
      cid__.clear();
      fs__.clear();
    }
    catch (TidError err) {HandleError(err);}
  }

  list<string>::iterator r;
  list<unsigned int>::iterator p, f;

  for(p=cid_.begin(), r=chnames_.begin(), f=fs_.begin(); 
      p!=cid_.end(); 
      p++,r++,f++,id++) {
    cidid[*p] = id;
    namecid[*r]=*p;
    events.push_back(ListTidEventsDef());
    eventFiles.push_back(ListString());
    eventFirstTime.push_back(VectorDouble());
    data.push_back(ArrayBool());
    FirstTime.push_back(0);
    FirstTimeDefined.push_back(0);
    LastTime.push_back(0);
    TimeWidth.push_back(0);
    NFreq.push_back((unsigned int)ceil(0.5*(double)(*f)*TimeResolution));
  }
}




void TidClass::InitCatalog () {

  map<string, unsigned int> namecid;

  list<string>::iterator p;
  list<unsigned int>::iterator q = cid_.begin();

  for(p=chnames_.begin(); p!=chnames_.end(); p++, q++) 
    namecid[*p] = *q;


  FILE *ifp;
  char comm[2048], buf[2048];
  list<string> filelist;

  /* Get list of files */
  sprintf(comm,"ls -1 %s/*.tid",catPath.c_str());
  if((ifp = popen(comm,"r")) == NULL)
    throw TidError (string("Unable to get directory content for ") + catPath);
 
  while(fscanf(ifp,"%s",buf)!=EOF)
    {
      if(strcmp(buf,"ls: No match.")==0)
	throw TidError (string("Empty catalog: ") + catPath);

      filelist.push_back(buf);
      
      catFiles.push_back(buf);

      struct stat fs;

      stat(buf,&fs);

      catTimes.push_back(fs.st_mtime);
    }
  pclose(ifp);

  if(!filelist.size())
    TidWarning (string("Empty catalog: ") + catPath);

  unsigned int GoodFiles=0;

  for(p=filelist.begin(); p!=filelist.end(); p++) {

    try {
      ReadCatFile(*p);
      TidMessage(string("File ") + *p + string(" successfully added to catalog.")); 
      GoodFiles++;
    }
    catch (TidError err) {
      HandleError(err);
      TidWarning(string("Skipping invalid catalog file ") + *p);
    }

  }

  if(!GoodFiles) 
    throw TidError (string("Empty catalog: ") + catPath);


  
  unsigned int i;
  list<string>::iterator ls;
  string catOut("Event Catalog:\nName\t\tChannel\t\tMax Duration(s)\n");
  for(i=0, 
	ls=chnames_.begin(); 
      i<events.size(); 
      i++, 
	ls++) 
    if(events[i].size()) {
      char buf[256];

      ListTidEventsDef::iterator p;

      for(p=events[i].begin(); p!=events[i].end(); p++) {
	if((*p)->rect) {

	  double TimeW =  TimeResolution + (*p)->rect->GetMaxT();

	  if(TimeW > TimeWidth[i]) TimeWidth[i] = TimeW;

	  sprintf(buf,"%g",TimeW);

	  catOut += (*p)->desc + string("\t\t") + *ls + string("\t\t") + string(buf) + string("\n");

	  
	}
      }

    }

  TidMessage(catOut);

}




void TidClass::ReadCatFile(string s) {
  /* format is:
     class Name
     channel=H0:PEM-MX_SEISX
     alpha=12
     rectangle(lotb,hitb,lofb,hifb)
     rectangle(lotb,hitb,lofb,hifb,lotw,hitw,lofw,hifw)
     ...
     var(name,lobnd,hibnd,#steps)
     ...
     end
  */

  FILE* in;
  char buf[2048];
  unsigned int lnumb = 0;

  if(!(in=fopen(s.c_str(),"r"))) 
    throw TidError(string("Unable to open file ") + s + string(" for reading."));

  double BPP=-1.0;
  bool InProgress=0;
  bool AlphaDefined=0;
  bool ChannelDefined=0;
  bool MaxCostDefined=0;
  bool Debug=0;
  bool DebugFile=0;
  bool NormalizeCost = 0;

  string dFile;
  TidEventsDef* tmpDef;
  double alpha, maxCost;
  string channel;
  list<string> var, lofreq, hifreq, nsteps;

  for(;;) {

    *buf=0;
    fgets(buf, 2048, in);

    lnumb++;

    if(buf[0]!='#' && buf[0]!='\n' && buf[0]!=0) { // check for comments or empty lines
      

      bool knownOption = 0;

      if(strstr(buf,"class")==buf) {

	knownOption=1;

	if(InProgress) 
	  throw TidError("Missing end or attempt to start a new class within another one.");

	InProgress=1;

	char *tmp = buf+5;
	while(*tmp==32) tmp++;

	char* ptr2 = strrchr(tmp,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';

	tmpDef = new TidEventsDef(string(tmp));
      }
      

      if(strstr(buf,"Debug")==buf) {
	knownOption=1;

	if(!InProgress)
	  throw TidError("Misplaced Debug statement.");

	char *pt1 = strchr(buf,'('),
	  *pt2 = strchr(buf,')');

	if(pt1) {
	  if(pt2 && pt2>pt1) {
	    *pt2=0;
	    dFile=string(pt1+1);
	    DebugFile=1;
	  }
	  else throw TidError("Malformed Debug statement.");
	}
	else if(pt2)
	  throw TidError("Malformed Debug statement.");

	Debug=1;
      }

      if(strstr(buf,"NormalizeCost")==buf) {
	knownOption=1;

	if(!InProgress)
	  throw TidError("Misplaced NormalizeCost statement.");

	char *pt1 = strchr(buf,'('),
	  *pt2 = strchr(buf,')');

	if(pt1) {
	  if(pt2 && pt2>pt1) {
	    *pt2=0;
	    BPP = atof(pt1+1);
	  }
	  else throw TidError("Malformed NormalizeCost statement.");
	}
	else if(pt2)
	  throw TidError("Malformed NormalizeCost statement.");


	NormalizeCost = 1;
      }

      if(strstr(buf,"channel")==buf) {
	knownOption=1;
	if(!InProgress)
	  throw TidError("Misplaced channel statement.");
	
	char *ptr = strchr(buf,'=');
	char* ptr2 = strrchr(buf,'\n');
	if(ptr2 != NULL) *ptr2 = '\x0';
	channel=string(ptr+1);

	ChannelDefined=1;
      }

      if(strstr(buf,"alpha")==buf) {
	knownOption=1;
	if(!InProgress)
	  throw TidError("Misplaced alpha statement.");

	char *ptr = strchr(buf,'=');
	alpha=atof(ptr+1);

	AlphaDefined=1;
      }


      if(strstr(buf,"NStop")==buf) {
	knownOption=1;
	if(!InProgress)
	  throw TidError("Misplaced NStop statement.");

	char *ptr = strchr(buf,'=');
	tmpDef->SetmaxAboveThrStop((unsigned int)floor(atof(ptr+1)/TimeResolution));

      }


      if(strstr(buf,"maxcost")==buf) {
	knownOption=1;
	if(!InProgress)
	  throw TidError("Misplaced maxcost statement.");

	char *ptr = strchr(buf,'=');
	maxCost=atof(ptr+1);

	MaxCostDefined=1;
      }

      if(strstr(buf,"var")==buf) {

	knownOption=1;
	if(!InProgress)
	  throw TidError("Misplaced var statement.");

	char *ptr = strchr(buf,'(');
	char *eptr = strrchr(buf,')');
	if(!ptr || !eptr) 
	  throw TidError ("Malformed var statement.");

	*eptr = 0;

	char *nam = ptr+1;

	ptr=strchr(nam,',');
	if(!ptr) throw TidError ("Malformed var statement.");
	*ptr = 0;

	char *lobnd = ptr+1;

	ptr=strchr(lobnd,',');
	if(!ptr) throw TidError ("Malformed var statement.");
	*ptr = 0;

	char *hibnd = ptr+1;

	ptr=strchr(hibnd,',');
	if(!ptr) throw TidError ("Malformed var statement.");
	*ptr = 0;

	char *nstep = ptr+1;

	var.push_back(string(nam));
	lofreq.push_back(string(lobnd));
	hifreq.push_back(string(hibnd));
	nsteps.push_back(string(nstep));
      }


      if(strstr(buf,"rectangle")==buf) {
	knownOption=1;
	if(!InProgress)
	  throw TidError("Misplaced rectangle statement.");
	if(!AlphaDefined)
	  throw TidError("alpha statement must be before rectangle statement.");


	if(!(tmpDef->rect)) tmpDef->rect=new TidCompositeRectangle();

	//CONSTRUCT RECTANGLE
	vector<string> params, variables;

	char *ptr = buf + 9; // points to first argument
	char *now;

	while((now=strtok(ptr,",()\n"))) {

	  ptr = NULL;

	  char* varname = strpbrk(now,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ");

	  if(varname) { // have a variable

	    list<string>::iterator n=var.begin(),l=lofreq.begin(),u=hifreq.begin(),s=nsteps.begin();

	    char varbuf[2];
	    varbuf[0]=*varname; varbuf[1]=0;

	    for(;n!=var.end();n++,l++,u++,s++)
	      if(*n == string(varbuf)) break;

	    if(n==var.end())
	      throw TidError (string("Variable ") + varbuf + string(" was not defined before it is called."));

	    params.push_back(string(now));
	    variables.push_back(*n);
	    variables.push_back(*l);
	    variables.push_back(*u);
	    variables.push_back(*s);
	  }
	  else { // purely numerics
	    params.push_back(string(now));
	  }
	}

       
	if(params.size()==8) {

	  for(unsigned int vi=0; vi<variables.size(); vi++)
	    params.push_back(variables[vi]);

	  TidRectangle tmpRectangle(TimeResolution,alpha,8,params);
	  //	  tmpRectangle.SetMaxT();
	  tmpDef->rect->PushBack(tmpRectangle);
	
	}
	else if(params.size()==4) {

	  throw TidError("Call to rectangle with only 4 parameters not yet implemented.");

	}
	else throw TidError("Invalid number of parameters.");

	params.clear();
	variables.clear();
	
      }

      if(strstr(buf,"end")==buf) {

	knownOption=1;

	if(!InProgress)
	  throw TidError("Misplaced end statement.");

	if(!AlphaDefined)
	  throw TidError("Missing alpha definition.");

	if(!ChannelDefined)
	  throw TidError("Missing channel name.");

	if(!MaxCostDefined)
	  throw TidError ("Missing maxcost definition.");
	
	if(Debug) {
	  if(tmpDef->rect) tmpDef->rect->SetDebug();
	}

	if(DebugFile) {
	  if(tmpDef->rect) tmpDef->rect->SetDebugFile(dFile);
	}

	if(NormalizeCost) tmpDef->NormalizeCost = 1;

	if(BPP>0.0) tmpDef->BPP = BPP; 

	map<string,unsigned int>::iterator ci = namecid.find(channel);

	if(ci==namecid.end()) {
	  TidWarning(string("Channel ") + channel + string(" is not being processed by any of the data server that are presently connected. This channel will be skipped."));
	}
	else {
	  tmpDef->maxCost = maxCost;
	  events[cidid[ci->second]].push_back(tmpDef);	  
	  eventFiles[cidid[ci->second]].push_back(s);
	  eventFirstTime[cidid[ci->second]].push_back(0.0);
	}
	
	NormalizeCost = 0;
	DebugFile = 0;
	Debug = 0;
	MaxCostDefined = 0;
	ChannelDefined = 0;
	AlphaDefined = 0;
	InProgress=0;
      }



      if(!knownOption) 
	throw TidError ("Unknown option.");
      
    }

    if(feof(in)) {
      if(InProgress) 
	throw TidError("Missing end statement.");
      else break;  
    }

  }

  fclose(in);

}


bool TidClass::Run () {

  void *sleep_(void *s);

  pthread_t sleeper;
  //  void * status;

  try {
    unsigned int SleepTime_ = SleepTime;

    if(pthread_create(&sleeper, NULL, sleep_, &SleepTime_))
      throw TidError("Thread creation error");

    // get the data, create pixmap
    GetData();

    // minimize cost & broadcast results
    Classify();

    RereadCatalog();

    if(pthread_join(sleeper, NULL)) // wait if time not elapsed
      throw TidError("Thread join error");
  }
  catch (TidError err) {HandleError(err);}

    return 1;
}


void *sleep_(void *s) {
  sleep(*((unsigned *)s));
  return NULL;
}


/*
// Debug version
void TidClass::GetData() {

  static double time = 0.0;

  list<TidIO*>::iterator tio_;
  unsigned int i;
  for(tio_ = tio.begin(), i=0; tio_ != tio.end(); tio_++, i++) {
    list<EventLW*> dat;    
    unsigned int GPSsecT;
    (*tio_)->ReadEventList(dat, &GPSsecT);
    dat.clear();
  }

  time+=1.0;


  for(unsigned int id=0; id<2; id++) {

    if(!FirstTimeDefined[id]) {
      LastTime[id] = FirstTime[id] = 0.0;
      FirstTimeDefined[id]=1;
    }

    vector<bool> zeroVector(NFreq[id],0); 

    if(time>=500 && time<680) {
      zeroVector[27] = 1;
      zeroVector[28] = 1;
      zeroVector[29] = 1;
      zeroVector[30] = 1;
      zeroVector[31] = 1;
      zeroVector[32] = 1;

      zeroVector[55] = 1;
      zeroVector[56] = 1;

      zeroVector[70] = 1;
      zeroVector[71] = 1;
      zeroVector[72] = 1;
    }

    data[id].push_back(zeroVector);
    LastTime[id]=time;

    //    cout<<data[id].size()<<endl;
  }
}
*/


void TidClass::GetData() {

  list<EventLW*> dat;    
  unsigned int GPSsecT;

  try {
    
    list<TidIO*>::iterator tio_;
    unsigned int i;
    for(tio_ = tio.begin(), i=0; tio_ != tio.end(); tio_++, i++) {
  
      (*tio_)->ReadEventList(dat, &GPSsecT);


      if(GPSsecT < UpTimeRef) UpTimeRef = GPSsecT;
      if(GPSsecT > UpTimeMax) UpTimeMax = GPSsecT;
      UpTime = UpTimeMax - UpTimeRef;

      list<EventLW*>::iterator p;
      for(p=dat.begin(); p!=dat.end();p++) {

	(*p)->cid += 1000 * i;

	unsigned int t00 = (*p)->tfp[0].t;
	unsigned int id = cidid[(*p)->cid];

	if(events[id].size()) {

	if(!FirstTimeDefined[id]) {
	  LastTime[id] = FirstTime[id] = (double)((*p)->GPSSec) + 1E-9*(double)((*p)->GPSNSec);
	  FirstTimeDefined[id]=1;
	}

	for(unsigned int n=0; n<(eventFirstTime[id]).size(); n++)
	    if((eventFirstTime[id])[n] < 1.0) (eventFirstTime[id])[n] = (double)((*p)->GPSSec) + 1E-9*(double)((*p)->GPSNSec);

	for(unsigned int j=0; j<(*p)->tfp.size(); j++) {

	  double time = (double)((*p)->GPSSec) + 1E-9*(double)((*p)->GPSNSec) + TimeResolution * ((double)((*p)->tfp[j].t) - (double)t00);

	  int listind = (int)floor((time-FirstTime[id])/TimeResolution);
	  unsigned int freqind = (*p)->tfp[j].f;

	  if(listind<0) {
	    if(FirstTime[id] - time <= GraceTime) {
	      vector<bool> zeroVector(NFreq[id],0); 
	      for(unsigned int k=0; k<(unsigned int)abs(listind); k++) 
		data[id].push_front(zeroVector);
	      (data[id].front())[freqind]=1;
	      FirstTime[id]-=TimeResolution*(double)abs(listind);
	      for(unsigned int n=0; n<(eventFirstTime[id]).size(); n++)
		(eventFirstTime[id])[n]-=TimeResolution*(double)abs(listind);
	    }
	    else {
	      /*
	      TidWarning("You're recieving points too old to be processed efficiently. I will increase the GraceTime by 10% and get back to you if that doesn't fix the problem.");
	      GraceTime *= 1.1;
	      char bufGT[256];
	      sprintf(bufGT,"%g",GraceTime);
	      TidWarning(string("The new GraceTime is ") + bufGT + string(" seconds."));
	      */
	      TidWarning("Recieving points too old to be processed efficiently. Skipping data!");
	      printf("%.15g\t\t%.15g\t\t%.15g\n",time,FirstTime[id],GraceTime);
	      //printf("%u\t%u\t\t%u\t%u\t\t%u\n",(*p)->GPSSec, (*p)->GPSNSec, (*p)->tfp[j].t, t00, freqind);
	    }
	  }
	  else if(listind >= (int)data[id].size()) {
	    vector<bool> zeroVector(NFreq[id],0); 
	    int tmpsiz = (int)data[id].size()-1;
	    for(int k=0; k<listind - tmpsiz; k++) 
	      data[id].push_back(zeroVector);
	    (data[id].back())[freqind]=1;
	    LastTime[id]+=TimeResolution*(double)(listind-tmpsiz);
	  }
	  else {
	    list<VectorBool>::iterator p;
	    int n;
	    for(p=data[id].begin(), n=0; n<listind; p++, n++) {}
	    (*p)[freqind]=1;
	  }
	 

	  //	  cout<<listind<<"\t\t"<<data[id].size()<<endl;
	}
	}
	delete (*p);
      }
      dat.clear();
    }
  }

  catch (TidError err) {HandleError(err);}
}



void TidClass::Classify () {

  for(unsigned int i=0; i<events.size(); i++) {
    //    printf("%.15g\t\t%.15g\n",FirstTime[i],LastTime[i]);
    if(events[i].size()) {

      // loop over catalog
      ListTidEventsDef::iterator ev;
      unsigned int j;
      bool oneGood = 1;

      /*
      while(oneGood) {
      */
      while(LastTime[i] - FirstTime[i] > GraceTime + TimeWidth[i]) {

	oneGood = 0;

	for(ev=events[i].begin(), j=0; ev!=events[i].end(); ev++, j++) {

	  //	   printf("%.15g\t%.15g\t\t%.15g\n",LastTime[i],(eventFirstTime[i])[j],GraceTime + TimeWidth[i]);
	  if(LastTime[i] - (eventFirstTime[i])[j] >= GraceTime + TimeWidth[i]) {
	    oneGood = 1;
	    double evDuration = -1.0;
	    vector<string> outPar;
	    
	    if((*ev)->rect->Test(&(data[i]),(*ev)->maxCost,&outPar,&evDuration,FirstTime[i], (*ev)->BPP, (*ev)->NormalizeCost)) {

	      (*ev)->GotTrigger();

	      // broadcast

	      if(DoTrigger) {
#ifdef WithDMT	    

		double timm = atof(outPar[1].c_str());

		trig::TrigBase trigData("TID",
					(*ev)->desc.c_str(),
					Time((long unsigned int)floor(timm), (long unsigned int)floor(1E9*(timm - floor(timm)))),
					Interval(evDuration),
					atof(outPar[0].c_str()));
		trigData.setPriority(trig::p_warn);    // Set severity level
		trigData.setDisposition(trig::d_metaDB);   // Route to meta-data
		gTidClass->SendTrig(trigData);

#else
		throw TidError("FATAL: trigger generation not supported!");
#endif
	      }

	      if(HtmlDef) {
#ifdef WithDMT
		try {
		  int row = htmlTable->addRow(); //-1;
		  string varList;
		  for(unsigned int j=2; j<outPar.size(); j+=2)
		    varList += outPar[j] + " = " + outPar[j+1] + "\t";
		  
		  htmlTable->insertData(row,0,(*ev)->desc);
		  htmlTable->insertData(row,1,outPar[1]);
		  htmlTable->insertData(row,2,outPar[0]);
		  htmlTable->insertData(row,3,varList);
		  
		  htmlDocu = new html::document("Tid Output");
		  //    htmlDocu->setBackgroundColor(color(1,1,1));
		  htmlDocu->setRefresh(60.0); // refresh every minute
		  htmlDocu->addObject(*htmlTable);
		  
		  htmlStream = new std::ofstream(htmlFile.c_str());
		  if(!(*htmlStream)) 
		    throw TidError ( string("Unable to open file ") + htmlFile );

		  htmlWriter = new html::writer(*htmlStream);
		  
		  htmlDocu->write(*htmlWriter);
		  htmlWriter->sync();

		  htmlStream->close();
		  delete htmlStream;
		  delete htmlWriter;
		  delete htmlDocu;
		  
		}
		catch (std::runtime_error err) {
		  throw TidError("Error in html file creation.");
		}
#endif
#ifndef WithDMT
		throw TidError("FATAL: Html output is not supported on this compilation. Make sure to edit the Makefile properly if you need Html output.");
#endif	      
	      }
	      
	      if(ASCIIDef) {
		
		(*ASCIIStream)<<(*ev)->desc<<"\t"<<outPar[1]<<"\t"<<outPar[0];
		for(unsigned int j=2; j<outPar.size(); j++)
		  (*ASCIIStream)<<"\t"<<outPar[j];
		(*ASCIIStream)<<"\n";
		
	      }
	      

	      printf("%s\t%s\t%s\n",(*ev)->desc.c_str(), outPar[1].c_str(),outPar[0].c_str());
	      for(unsigned int j=2; j<outPar.size(); j++)
		printf("%s\t",outPar[j].c_str());
	      printf("\n");
	 
	    }
	    
	    outPar.clear();
	  } // end time check
	} // end loop on event[i]

	if(oneGood) {
	  if(data[i].size()) {
	    //	    cout<<data[i].size()<<endl;
	    data[i].front().clear();
	    data[i].pop_front();
	    FirstTime[i] += TimeResolution;

	    /*	    
	    for(unsigned int j=0; j<eventFirstTime[i].size(); j++)
	      if(LastTime[i] - (eventFirstTime[i])[j] >= GraceTime + TimeWidth[i])
		(eventFirstTime[i])[j] += TimeResolution;
	    */
	  }
	}

      }// end oneGood

    } // end if(evnet[i].size())
      

  } // end loop on i
}


void TidClass::RereadCatalog () {

  static unsigned int Slepper = 0;

  Slepper++;

  if(Slepper >= CatRefreshTime) {

  Slepper = 0;

  typedef list<string> ListString;

  FILE *ifp;
  char comm[2048], buf[2048];
  ListString filelist;
  ListString::iterator ibv;
  for(ibv = catFiles.begin(); ibv != catFiles.end(); ibv++)
    filelist.push_back(*ibv);

  list<string> newFiles;
  list<string> modFiles;
  list<string> dontDeleteFiles, delFiles;

  /* Get list of files */
  sprintf(comm,"ls -1 %s/*.tid",catPath.c_str());

  if((ifp = popen(comm,"r")) == NULL) 
    throw TidError (string("Unable to obtain directory content for ") + catPath);

	unsigned int nfiless = 0;

  while(fscanf(ifp,"%s",buf)!=EOF)
    {
      if(strcmp(buf,"ls: No match.")==0)
	throw TidError (string("Empty catalog: ") + catPath);

      struct stat fs;

      if(stat(buf,&fs)==-1) {
	perror("System error: ");
	throw TidError ("Cannot stat file!\n");
	}

	nfiless++;

      list<string>::iterator fn;
      list<time_t>::iterator ct;
      unsigned int fmod;

      for(fn=catFiles.begin(), fmod = 0, ct=catTimes.begin(); 
	  fn!=catFiles.end(); 
	  fn++, fmod++, ct++) 
	if((*fn) == string(buf)) break;

	if(fn == catFiles.end()) { // new file
	cout<<"New file: "<<buf<<endl;
	  catFiles.push_back(buf);
	  catTimes.push_back(fs.st_mtime);	  
	  newFiles.push_back(buf);
	  dontDeleteFiles.push_back(buf);
	}
	else {
	  if(fs.st_mtime != *ct) { // modified file
	cout<<"File "<<buf<<" was modified; old time: "<<*ct<<", new time: "<<fs.st_mtime<<endl;
	    *ct = fs.st_mtime;
	    modFiles.push_back(buf);
	    dontDeleteFiles.push_back(buf);
	  }
	  else {
	    dontDeleteFiles.push_back(buf);
	  }
	}

      
    }
  pclose(ifp);

if(!nfiless) throw TidError ("Unable to get file list!");
 
  if(!filelist.size())
    throw TidError (string("Empty catalog: ") + catPath);

  delFiles.clear();

  ListString::iterator p,pp;
/*
cout<<"**********************************************\n";
for(p=filelist.begin(); p!=filelist.end(); p++)
	cout<<*p<<endl;
cout<<endl;

for(pp=dontDeleteFiles.begin(); pp!=dontDeleteFiles.end(); pp++)
	cout<<*pp<<endl;
*/
  for(p=filelist.begin(); p!=filelist.end(); p++) {
    for(pp=dontDeleteFiles.begin(); pp!=dontDeleteFiles.end(); pp++)
      {
//	cout<<*p<<endl;
//	cout<<*pp<<endl;
	if((*pp) == (*p)) {
//	cout<<"same\n\n";
	break;
	}
      } 
    if(pp == dontDeleteFiles.end()) {
cout<<"deleting file "<<*p<<endl;
      delFiles.push_back(*p);
    }
  }

  for(p = modFiles.begin(); p != modFiles.end(); p++) {
    delFiles.push_back(*p);
    newFiles.push_back(*p);
  }

  /*********************************************************************/

  if(delFiles.size() || modFiles.size() || newFiles.size()) 
    TidMessage("The catalog was modified. Rereading.");
  else return;


  // delete patterns from deleted files
  for(p=delFiles.begin(); p!=delFiles.end(); p++) {

    unsigned int cid;
    ListString::iterator q;
    ListTidEventsDef::iterator q1;
    VectorDouble::iterator vd;

    for(cid=0;cid<events.size();cid++)
      for(q=eventFiles[cid].begin(), q1=events[cid].begin(), vd=eventFirstTime[cid].begin(); 
	  q!=eventFiles[cid].end(); 
	  q++, q1++, vd++)
	if(*q == *p) {
	  TidMessage(string("Deleting event associated to file ") + *p);

	  eventFiles[cid].erase(q);
	  events[cid].erase(q1);
	  eventFirstTime[cid].erase(vd);

	  list<string>::iterator fn;
	  list<time_t>::iterator ct;

	  for(ct=catTimes.begin(), fn=catFiles.begin(); fn != catFiles.end(); fn++, ct++) 
	    if(*fn == *p) break;
	  if(fn!=catFiles.end()) {
	    catFiles.erase(fn);
	    catTimes.erase(ct);
	  }

	  break;
	}

  }



  // (re)load relevant files
  for(p=newFiles.begin(); p!=newFiles.end(); p++) {

    try {
      ReadCatFile(*p);
      TidMessage(string("File ") + *p + string(" successfully added to catalog.")); 
    }
    catch (TidError err) {
      HandleError(err);
      TidWarning(string("Skipping invalid catalog file ") + *p);
    }

  }

  /*********************************************************************/

  unsigned int i;
  list<string>::iterator ls;
  string catOut("Event Catalog:\nName\t\tChannel\t\tMax Duration(s)\n");
  for(i=0, 
	ls=chnames_.begin(); 
      i<events.size(); 
      i++, 
	ls++) 
    if(events[i].size()) {
      char buf[256];

      ListTidEventsDef::iterator p;

      for(p=events[i].begin(); p!=events[i].end(); p++) {
	if((*p)->rect) {

	  double TimeW =  (*p)->rect->GetMaxT();

	  if(TimeW > TimeWidth[i]) TimeWidth[i] = TimeW;

	  sprintf(buf,"%g",TimeW);

	  catOut += (*p)->desc + string("\t\t") + *ls + string("\t\t") + string(buf) + string("\n");

	  
	}
      }
    }

  TidMessage(catOut);

newFiles.clear();
  modFiles.clear();
  dontDeleteFiles.clear();
  delFiles.clear();
  }
}



void TidClass::Report () {

  
#ifndef WithDMT
  throw TidError ("FATAL: Status Report no supported because code is not linked to a valid DMT distribution.");
#endif


  try {
    html::table *etable = new html::table("Running Patterns");
    etable->addColumn("Class name");
    etable->addColumn("Channel");
    etable->addColumn("Duration");
    etable->addColumn("Number found");
    etable->setBorder(1);

    unsigned int i;
    list<string>::iterator ls;
    for(i=0, 
	  ls=chnames_.begin(); 
	i<events.size(); 
	i++, 
	ls++) 
      if(events[i].size()) {

	char buf[256];

	ListTidEventsDef::iterator p;

	for(p=events[i].begin(); p!=events[i].end(); p++) {
	  if((*p)->rect) {

	    double TimeW =  (*p)->rect->GetMaxT();

	    sprintf(buf,"%g",TimeW);

	    int row = etable->addRow();

	    etable->insertData(row,0,(*p)->desc);
	    etable->insertData(row,1,*ls);
	    etable->insertData(row,2,string(buf));

	    sprintf(buf,"%u",(*p)->NTrigger);
	    etable->insertData(row,3,string(buf));
	  }
	}
      }

    html::document * htmlDocul = new html::document("Tid Event Catalog Status");
    htmlDocul->setRefresh(60.0); // refresh every minute

    string serv=source_.front();
    list<string>::iterator lsi;
    
    for(lsi=source_.begin(), lsi++;lsi!=source_.end();lsi++)
      serv += string(", ") + *lsi;

    htmlDocul->addObject(html::text(string("Data Server(s): ") + serv));

    htmlDocul->lineBreak();
    
    htmlDocul->addObject(html::text(string("Catalog Path: ") + catPath));

    htmlDocul->lineBreak();
    htmlDocul->lineBreak();
    htmlDocul->lineBreak();


    if(HtmlDef) htmlDocul->addObject(html::text(string("HTML trigger generation: ON")));
    else htmlDocul->addObject(html::text(string("HTML trigger generation: OFF")));

    htmlDocul->lineBreak();

    if(ASCIIDef) htmlDocul->addObject(html::text(string("ASCII trigger generation: ON")));
    else htmlDocul->addObject(html::text(string("ASCII trigger generation: OFF")));

    htmlDocul->lineBreak();

    if(DoTrigger) htmlDocul->addObject(html::text(string("LDAS trigger generation: ON")));
    else htmlDocul->addObject(html::text(string("LDAS trigger generation: OFF")));

    htmlDocul->lineBreak();
    htmlDocul->lineBreak();
    htmlDocul->lineBreak();

    char buf[256];
    sprintf(buf,"%u",UpTime);
    htmlDocul->addObject(html::text(string("Uptime: ") + string(buf) + string(" seconds")));

    htmlDocul->lineBreak();
    htmlDocul->lineBreak();
    htmlDocul->lineBreak();

    htmlDocul->addObject(*etable);


    std::ofstream * htmlStreaml = new std::ofstream(ReportFile.c_str());
    if(!htmlStreaml) 
      throw TidError ( string("Unable to open file ") + ReportFile );

    html::writer * htmlWriterl = new html::writer(*htmlStreaml);
		  
    htmlDocul->write(*htmlWriterl);
    htmlWriterl->sync();

    htmlStreaml->close();
    delete htmlStreaml;
    delete htmlWriterl;

    delete htmlDocul;
    delete etable;
  }
  catch (std::runtime_error err) {
    throw TidError("Error in html Status Report creation.");
  }

}


void TidClass::DumpCat (const char *file) {

  FILE *ou;

  if((ou=fopen(file,"w"))==NULL) {
    cerr<<"Unable to open "<<file<<endl;
    exit(-1);
  }

  

  unsigned int i;
  list<string>::iterator ls;
  
  for(i=0, 
	ls=chnames_.begin(); 
      i<events.size(); 
      i++, 
	ls++) 
    if(events[i].size()) {

      ListTidEventsDef::iterator p;

      for(p=events[i].begin(); p!=events[i].end(); p++) {

	fprintf(ou,"%s\n",(*p)->desc.c_str());

	if((*p)->rect) {

	  (*p)->rect->Dump(ou);
	  
	}
      }
    }



  fclose(ou);
}
